// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// Licensed under the MIT License:
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

var fs = require("fs");
var capnp = require("./capnp");
var assert = require("assert");
var Promise = require("es6-promise").Promise;
var spawn = require("child_process").spawn;

var goldenBinary = fs.readFileSync("src/node-capnp/testdata/binary");

var test = require("./test.capnp");
assert(test === capnp.import(__dirname + "/test.capnp"));
assert(test === require("./test"));

assert("namespace" in capnp.importSystem("capnp/c++.capnp"));

var parsed = capnp.parse(test.TestAllTypes, goldenBinary);

var roundTripped = capnp.serialize(test.TestAllTypes, parsed);

var canon = capnp.bytesToPreorder(test.TestAllTypes, roundTripped);

assert.equal(goldenBinary.length, roundTripped.length, "Round trip changed size?");
assert.equal(goldenBinary.toString("base64"), canon.toString("base64"), "Round trip lost data?");

// =======================================================================================

/* TODO(someday):  Revive this test.

var Fiber = require("fibers");

function wait(promise) {
  var fiber = Fiber.current;
  var success, result, error;
  promise.then(function (p) {
    success = true;
    result = p;
    fiber.run();
  }, function (e) {
    success = false;
    error = e;
    fiber.run();
  });
  Fiber.yield();
  if (success) {
    return result;
  } else {
    throw error;
  }
}

function doFiber(func, child) {
  new Fiber(function () {
    try {
      func();
      if (child) {
        child.kill();
        child.unref();
      }
    } catch (err) {
      console.log(err.stack);
      if (child) {
        child.kill();
        child.unref();
      }
      process.exit(1);
    }
  }).run();
}

var child = spawn("capnp-samples/calculator-server", ["127.0.0.1:21311"],
                  {stdio: [0, "pipe", 2], env: {}});

child.stdio[1].once("readable", function() {
  child.stdio[1].resume();  // ignore all input

  doFiber(function() {
    var conn = capnp.connect("127.0.0.1:21311");
    var Calculator = capnp.import("capnp-samples/calculator.capnp").Calculator;
    var calc = conn.restore("calculator", Calculator);

    var add = calc.getOperator("add").func;
    var subtract = calc.getOperator("subtract").func;
    var pow = {
      call: function (params) {
        return Math.pow(params[0], params[1]);
      }
    };

    var localCap = new capnp.Capability(pow, Calculator.Function);
    assert.equal(9, wait(localCap.call([3, 2])).value);
    localCap.close();

    var promise = calc.evaluate(
        {call: {"function": subtract, params: [
            {call: {"function": add, params: [
                {literal: 123}, {literal: 456}]}},
            {literal: 321}]}});

    var value = promise.value;
    assert.equal(258, wait(value.read()).value);
    value.close();

    value = calc.evaluate(
        {call: {"function": pow, params: [{literal: 2}, {literal: 4}]}}).value;
    assert.equal(16, wait(value.read()).value);
    value.close();

    // Try wrapping a promise as a capability -- calls are queued until resolution.
    var resolvePromisedCalc;
    var promisedCalc = new capnp.Capability(new Promise(function (resolve, reject) {
      resolvePromisedCalc = resolve;
    }), Calculator);

    value = promisedCalc.evaluate(
        {call: {"function": add, params: [
            {literal: 123}, {literal: 321}]}}).value;
    promise = value.read();
    resolvePromisedCalc(calc);
    assert.equal(444, wait(promise).value);
    value.close();

    // Like above, but reject the promise so queued calls fail.
    var rejectPromisedCalc;
    promisedCalc = new capnp.Capability(new Promise(function (resolve, reject) {
      rejectPromisedCalc = reject;
    }), Calculator);

    value = promisedCalc.evaluate(
        {call: {"function": add, params: [
            {literal: 123}, {literal: 321}]}}).value;
    promise = value.read();
    rejectPromisedCalc(new Error("foo example error"));
    assert.throws(function() { wait(promise); }, /foo example error/);
    value.close();

    add.close();
    subtract.close();
    conn.close();

    console.log("rpc: pass");
  }, child);
});
*/

console.log("pass");

